/* duplextest.c : Ethernet duplex-mismatch tester
 *
 * This program uses ICMP ECHO packets differently than other
 * standard tools in that it sends multiple packets before listening
 * for responses.  This attempts to create multiple packets "in
 * flight" before the first response is received, increasing the
 * chance of a duplex mismatch along the wire causing one or more of
 * the packets to be dropped.
 *
 * "ping -f" and other ping-like utilities will usually not see
 * duplex problems because they will usually only send one packet at
 * a time.  As the response is not sent until the packet is fully
 * received, there is usually no time when the wire is used in both
 * directions at the same time unless other traffic is already on
 * the wire.
 *
 * THIS PROGRAM IS NOT COMPLETELY RELIABLE.  Dropped packets are
 * always an indication of a problem, but no dropped packets does
 * not necessarily indicate that the link is working properly.  The
 * multiple packet algorithm was developed from a theory and quickly
 * refined by experimentation, but is likely not the best possible
 * implementation.  Future versions of this program may provide
 * better accuracy.  This program relies heavily on timing conditions
 * and will usually only work when the sender and receiver are on the
 * same LAN.
 *
 * This program can optionally be made setuid root to allow execution
 * by non-root users.  Privileges are dropped as soon as the socket
 * is acquired.
 *
 * Simon Kirby <sim@netnation.com>, 2002/02/12
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/param.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <sched.h>
#include <netdb.h>
#include <sys/poll.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>

#define PING_VERBOSE_DROP 0x0001
#define PING_VERBOSE_RECV 0x0002

#define MAX_RECV_SIZE 1600

static int open_socket_icmp(){
	int fd;

	fd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (fd != -1)
		return fd;

	fprintf(stderr,"Could not open ICMP socket: %s\n",strerror(errno));
	exit(1);
}

/*
 * Checksum algorithm stolen from "ping.c".
 * Original license from "ping.c":
 *
 * Copyright (c) 1989 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Muuss.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static unsigned short in_cksum(const u_short *addr,register int len,u_short csum){
	register int nleft = len;
	const u_short *w = addr; 
	register u_short answer; 
	register int sum = csum; 

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while (nleft > 1){
		sum += *w++;
		nleft -= 2; 
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1)
		sum += htons(*(unsigned char *)w << 8);

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}

static int send_icmp_ping(int fd,unsigned int payload_length){
	struct icmp *icp;
	unsigned int length;
	int r;
	struct {
		struct icmp icmp;
		unsigned char payload[payload_length];
	} packet;
	static int id = 0;

	length = sizeof(struct icmp) + payload_length;
	memset(&packet,0,length);
	memset(&packet.payload,0x55,payload_length);

	icp = &packet.icmp;
	icp->icmp_type = ICMP_ECHO;
	icp->icmp_code = 0;
	icp->icmp_cksum = 0;
	icp->icmp_hun.ih_idseq.icd_id = ((unsigned int)getpid()) & 0xffff;
	icp->icmp_hun.ih_idseq.icd_seq = ++id;

	icp->icmp_cksum = in_cksum((u_short *)icp,length,0);

	r = send(fd,&packet,length,0);
	if (r == length)
		return 0;
	if (r == -1)
		perror("sendto()");
	else
		fprintf(stderr,"sendto(): Short write\n");
	return -1;
}

static int recv_icmp_ping(int fd,struct sockaddr_in *ip,int timeout_msec,int flags){
	int r;
	socklen_t slen;
	struct timeval tv;
	struct sockaddr_in from;
	struct {
		struct ip ip;
		struct icmp icmp;
		unsigned char payload[MAX_RECV_SIZE];
	} packet;
	unsigned short checksum;

	slen = sizeof(from);

	tv.tv_sec = timeout_msec / 1000;
	tv.tv_usec = (timeout_msec % 1000) * 1000;
	if (setsockopt(fd,SOL_SOCKET,SO_RCVTIMEO,(char *)&tv,sizeof(tv)) != 0)
		perror("setsockopt(SOL_SOCKET,SO_RCVTIMEO)");

	for (;;){
		r = recvfrom(fd,&packet,MAX_RECV_SIZE,0,(struct sockaddr *)&from,&slen);
		if (r > 0){
			if (memcmp(&from.sin_addr,&ip->sin_addr,sizeof(ip->sin_addr)) != 0)
				continue;
			r-= sizeof(struct ip);
			if (r < 0){
				fprintf(stderr,"Received undersized response.\n");
				if (flags & PING_VERBOSE_DROP)
					printf("<");
				return 0;
				continue;
			}
			if (r == 0){
				fprintf(stderr,"Received response with no data.\n");
				if (flags & PING_VERBOSE_DROP)
					printf("0");
				return 0;
			}
			if ((checksum = in_cksum((u_short *)&packet.icmp,r,0))){
				fprintf(stderr,"Received response with incorrect checksum (%04x).\n",
					checksum);
				if (flags & PING_VERBOSE_DROP)
					printf("X");
				return 0;
			}
			if (flags & PING_VERBOSE_RECV)
				printf("!");
			return r;
		}
		if (r == -1){
			if ((errno == EAGAIN) || (errno == EINTR)){
				if (flags & PING_VERBOSE_DROP)
					printf(".");
				return 0;
			}
			perror("recvfrom()");
			if ((errno != EBADF) && (errno != ENOTCONN) &&
			    (errno != ENOTSOCK) && (errno != EFAULT) &&
			    (errno != EINVAL))
				continue;
		}
		if (r == 0)
			fprintf(stderr,"Socket closed on receive?\n");
		exit(1);
	}
}

static void usage(const char *prog){
	fprintf(stderr,"duplextest by Simon Kirby <sim@netnation.com>\n\n"
		"Usage: %s [<options>] <IP>[/<CIDR netmask>]\n\n"
		"Options:\n"
		"	-v	Verbose mode (show bangs for received packets).\n"
		"	-q	Quiet mode (no extraneous information).\n"
		"	-s	Silent mode (only report problems, implies -q).\n"
		"	-h	This help.\n",
		prog);
}

int main(int argc,char *argv[]){
	struct hostent *hp;
	char *host,*c;
	int fd = open_socket_icmp();
	int i,p,r;
	int count_dropped,count_received;
	int verbose,quiet,silent,problem;
	struct sockaddr_in dest;
	struct in_addr ip,ip_end;
	struct timeval tv;

	setuid(getuid());	/* Drop privileges */

	/*
	 * Parse the command line.
	 */

	verbose = quiet = silent = 0;
	for (;;){
		i = getopt(argc,argv,"vsq");
		if (i == -1)
			break;
		switch (i){   
		case 'v':
			verbose++;
			break;
		case 's':
			silent++;
			/* Fallthrough */
		case 'q':
			quiet++;
			break;
		case 'h':
			usage(argv[0]);
			exit(0);
			break;
		default:
			fprintf(stderr,"Unrecognized option '%c'.\n",i);
			usage(argv[0]);
			exit(1);
		}
	}

	/*
	 * Parse input and resolve hosts.
	 */

	if (!argv[optind]){
		usage(argv[0]);
		exit(1);
	}

	host = strdup(argv[optind]);
	if (!host){
		perror("strdup()");
		exit(1);
	}
	if ((c = strchr(host,'/')))
		*c++ = '\0';

	memset(&dest,0,sizeof(dest));
	dest.sin_family = AF_INET;
	if (inet_aton(host,&dest.sin_addr) != 1){
		if (!quiet)
			fprintf(stderr,"Resolving \"%s\"...\r",host);
		hp = gethostbyname(host);
		if (!hp){
			fprintf(stderr,"Could not resolve \"%s\".\n",host);
			exit(1);
		}
		memcpy(&dest.sin_addr,hp->h_addr,sizeof(dest.sin_addr));
		if (!quiet)
			fprintf(stderr,"Resolved \"%s\" to \"%s\".\n",host,inet_ntoa(dest.sin_addr));
	}
	ip.s_addr = ip_end.s_addr = ntohl(dest.sin_addr.s_addr);

	if (c){
		i = atoi(c);
		ip_end.s_addr|= (1 << (32 - i)) - 1;
		if (verbose){
			dest.sin_addr.s_addr = htonl(ip.s_addr);
			fprintf(stderr,"Sweeping from \"%s\"",inet_ntoa(dest.sin_addr));
			dest.sin_addr.s_addr = htonl(ip_end.s_addr);
			fprintf(stderr," to \"%s\"...\n",inet_ntoa(dest.sin_addr));
		}
	}

#ifdef SO_SNDTIMEO
	/*
	 * Set send timeout to one second (wait one second if send queue
	 * is clogged).
	 */
	tv.tv_sec = 1;
	tv.tv_usec = 0;
	if (setsockopt(fd,SOL_SOCKET,SO_SNDTIMEO,(char *)&tv,sizeof(tv)) != 0)
		if (!quiet)
			fprintf(stderr,"Warning: setsockopt(SOL_SOCKET,SO_SNDTIMEO)");
#endif

#ifdef SOL_IP
	i = 1;	/* Enable receiving of errors such as ICMP unreachable, etc. */
	if (setsockopt(fd,SOL_IP,IP_RECVERR,(char *)&i,sizeof(i)) != 0)
		if (!quiet)
			fprintf(stderr,"Warning: Could not enable IP_RECVERR.\n");
#endif

	i = 1;	/* Enable address reuse */
	if (setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,(char *)&i,sizeof(i)) != 0)
		if (!quiet)
			fprintf(stderr,"Warning: Could not enable IP_RECVERR.\n");

	/* Set large send/receive buffers. */
	i = 65536;
	if (setsockopt(fd,SOL_SOCKET,SO_RCVBUF,(char *)&i,sizeof(i)) != 0){
		perror("setsockopt(SOL_SOCKET,SO_RCVBUF)");
	}
	i = 65536;
	if (setsockopt(fd,SOL_SOCKET,SO_SNDBUF,(char *)&i,sizeof(i)) != 0){
		perror("setsockopt(SOL_SOCKET,SO_SNDBUF)");
	}

	/*
	 * Ping to make sure the host is reachable and not firewalled.
	 * Also attempt to determine number of hops.
	 */

	fflush(stdout);
	setvbuf(stdout,NULL,_IONBF,0);

	problem = 0;

	for (;ip.s_addr <= ip_end.s_addr;ip.s_addr++){
		memset(&dest,0,sizeof(dest));
		dest.sin_family = AF_INET;
		dest.sin_addr.s_addr = htonl(ip.s_addr);

		if (!silent)
			printf("%-15s ",inet_ntoa(dest.sin_addr));

		if (connect(fd,(struct sockaddr *)&dest,sizeof(dest)) == -1){
			switch (errno){
			case EACCES:
				if (!silent)
					printf("Broadcast.\n");
				continue;
			default:
				perror("connect()");
				exit(1);
			}
		}

		send_icmp_ping(fd,8);
		recv_icmp_ping(fd,&dest,100,0);

		count_dropped = count_received = 0;
		usleep(1);

		for (i = 0;i < 96;i++){
			send_icmp_ping(fd,8);
			send_icmp_ping(fd,12);
			send_icmp_ping(fd,24);
			send_icmp_ping(fd,48);
			for (p = 0;p < 4;p++){
				r = recv_icmp_ping(fd,&dest,50,
					((silent) ? 0 : PING_VERBOSE_DROP) |
					((verbose) ? PING_VERBOSE_RECV : 0));
				if (r > 0){
					if (!count_received++)
						count_dropped = 0;
					continue;
				}

				count_dropped++;
				if (count_received && count_dropped > 4){
					if (!silent)
						printf("Problem! (%u dropped in %u received)\n",
							count_dropped,count_received);
					problem = 1;
					goto next;
				}

				if (count_dropped > 16){
					if (!silent)
						printf("Unreachable.\n");
					goto next;
				}
			}
		}
		if (!silent)
			printf("Good.\n");
	next:
	}

	exit(problem);
}
